/* GIO - GLib Input, Output and Streaming Library
 *.
 * Copyright (C) 2010 Red Hat, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 * Author: Tomas Bzatek <tbzatek@redhat.com>
 */

#ifndef __gvfs_enums_h__
#define __gvfs_enums_h__

typedef enum
{
  G_VFS_BACKEND_NETWORK_DISPLAY_MODE_DISABLED,
  G_VFS_BACKEND_NETWORK_DISPLAY_MODE_SEPARATE,
  G_VFS_BACKEND_NETWORK_DISPLAY_MODE_MERGED
} GVfsBackendNetworkDisplayMode;

#endif /* __gvfs_enums_h__ */
